/*
****************************************************************************

              Copyright 1991, by FUJITSU LIMITED
              Copyright 1991, by Sun Microsystems, Inc.	      

Permission to use, copy, modify, distribute, and sell this software
and its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of FUJITSU LIMITED and Sun
Microsystems, Inc. not be used in advertising or publicity pertaining
to distribution of the software without specific, written prior
permission.  FUJITSU LIMITED and Sun Microsystems, Inc. makes no
representations about the suitability of this software for any
purpose.  It is provided "as is" without express or implied warranty.

FUJITSU LIMITED AND SUN MICROSYSTEMS, INC. DISCLAIMS ALL WARRANTIES
WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL FUJITSU LIMITED AND SUN
MICROSYSTEMS, INC. BE LIABLE FOR ANY SPECIAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

Auther: Riki	Kawaguchi       FUJITSU LIMITED
	A.	Yukisita        FUJITSU LIMITED
	Hideki  Hiura           (hhiura@Sun.COM) Sun Microsystems, Inc.

****************************************************************************
*/
#ifndef lint
#ident  "@(#)HttPreedit.c	1.17    93/02/12 SMI"
#endif

#ifdef X_LOCALE
#include <X11/Xlocale.h>
#else
#include <locale.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#if XlibSpecificationRelease != 5
#include <X11/XlibR5.h>
#endif /* XlibSpecificationRelease */
#include "XSunExt.h"

#include <X11/Xutil.h>
#include <X11/Xresource.h>

#include <X11/keysymdef.h>
#ifdef SVR4
#include <limits.h>
#endif
#include "Htt.h"
#include "XSunIMProt.h"
#include "HttXsfSun.h"

#ifndef MB_LEN_MAX
#define MB_LEN_MAX 4
#endif
#ifndef WCHAR_SIZE
#define WCHAR_SIZE 2
#endif

extern void	Display_ImlStatus_wcs() ;
extern htt_t	*htt;

static void	xsf_replace_feedback();
static void	xsf_replace_text();
static void	xsf_preedit_clear();
static void	xsf_preedit_erase();
static int	xsf_mbstowcs();

/*
 * Drawing Functions...
 *	XsfPreeditDraw,			XsfStatusDraw,
 */
extern void
XsfPreeditDraw( ic, pds )
    FXimp_Client		 *ic ;
    XIMPreeditDrawCallbackStruct *pds ;
{
    wchar_t	*preedit_string ;
    int		 preedit_length ;
    int         mode;
    FXimp_Client *focus_in_client=GET_HTT_VALUE(htt,focus_in_client);
    int preedit_feedback=GET_HTT_VALUE(htt,preedit_feedback);

#ifdef	DEBUG
    dbg_printf("*** Preedit Draw Server\n");
    dbg_printf("	pds->caret = %d\n", pds->caret);
    dbg_printf("	pds->chg_first = %d\n", pds->chg_first);
    dbg_printf("	pds->chg_length = %d\n", pds->chg_length);
#endif
    ic->preedit_text.caret_pos = pds->caret;
    ic->preedit_text.chg_first = pds->chg_first;

    if (preedit_feedback == FEEDBACK_STIPPLE) {
        if (ic == focus_in_client)
            mode = highlight;
        else
            mode = stipple;
    }
    else
        mode = highlight;

    /*
     * Preedit area clear
     */
    if ( ( !pds->chg_first )
      && ( !pds->chg_length )
      && ( !pds->text ) ) {
	xsf_preedit_clear( ic ) ;
    	DisplayOutputBuff(ic,mode,False) ;
    	DisplayCursor(ic, False) ;
    	return ;
    }

    /*
     * Preedit area erase
     */
    if ( !pds->text ) {
	xsf_preedit_erase( ic, pds->chg_first, pds->chg_length ) ;
    	DisplayOutputBuff(ic,mode,False) ;
    	DisplayCursor(ic, False) ;
    	return ;
    }

    /*
     * Preedit area erase
     */
    if ( ( !pds->text->length )
      && ( !pds->text->feedback )
      && ( !pds->text->string.wide_char ) ) {
	xsf_preedit_erase( ic, pds->chg_first, pds->chg_length ) ;
    	DisplayOutputBuff(ic,mode,False) ;
    	DisplayCursor(ic, False) ;
    	return;
    }

    /*
     * Text Infomation replace
     */
    xsf_replace_feedback( ic, pds );
    xsf_replace_text( ic, pds );
    DisplayOutputBuff(ic,mode,False) ;
    DisplayCursor(ic, False) ;
#ifdef	DEBUG
    dbg_printf("	preedit_char_len=%d\n",
				ic->preedit_text.preedit_char_len);
#endif
    return;
}

extern void
xsf_preedit_caret_server( ic, rv )
    FXimp_Client	*ic ;
    iml_inst		*rv ;	/* Return values from IMLogic() to be
				 *		interpreted */
{
    /*
     * Not Yet Fixed
     */
    return;
}

extern void
xsf_status_draw_server( ic, rv )
    FXimp_Client	*ic ;
    iml_inst		*rv ;	/* Return values from IMLogic() to be
							 *  interpreted */
{
    XIMStatusDrawCallbackStruct		*sds ;
    int mode;
    FXimp_Client *focus_in_client=GET_HTT_VALUE(htt, focus_in_client);
    int preedit_feedback=GET_HTT_VALUE(htt,preedit_feedback);

    if (preedit_feedback == FEEDBACK_STIPPLE) {
        if (ic == focus_in_client)
            mode = highlight;
        else
            mode = stipple;
    }
    else
        mode = highlight;

    sds = (XIMStatusDrawCallbackStruct *)&rv->operand ;
    if ( ic->status_message ) {
	free( ic->status_message ) ;
	ic->status_message = NULL ;
    }
    ic->status_length =
	xsf_mbstowcs( sds->data.text, &ic->status_message ) ;
    DisplayStatus_wcs( ic,mode ) ;
    return;
}

/*
 * Clear text
 */
static void
xsf_preedit_clear( ic )
    FXimp_Client    *ic ;
{
    ic->preedit_text.caret_pos = 0;
    ic->preedit_text.preedit_char_len = 0;

    if ( ic->preedit_text.preedit_char )
	free( ic->preedit_text.preedit_char );
    ic->preedit_text.preedit_char = NULL;
    if ( ic->preedit_text.preedit_attr )
	free( ic->preedit_text.preedit_attr );
    ic->preedit_text.preedit_attr = NULL;

    return;
}

/*
 * Erase text
 */
static void
xsf_preedit_erase( ic, erase_first, erase_length )
    FXimp_Client    *ic ;
    int		     erase_first;
    int		     erase_length;
{
    int			 length;
    Htt_PreeditTextRec	*txinfo = &ic->preedit_text;
    XIMFeedback		*fb1, *fb2;
    wchar_t		*wc1, *wc2;

    if ( erase_length == 0 )
	return;
    if ( ( !txinfo->preedit_attr )
      || ( !txinfo->preedit_char )
      || ( txinfo->preedit_char_len == erase_length ) ) {
	if ( txinfo->preedit_attr )
	    free( txinfo->preedit_attr );
	txinfo->preedit_attr = NULL;
	if ( txinfo->preedit_char )
	    free( txinfo->preedit_char );
	txinfo->preedit_char = NULL;

	txinfo->caret_pos = 0;
	txinfo->preedit_char_len = 0;
	return;
    }

    fb1 = txinfo->preedit_attr + erase_first;
    fb2 = fb1 + erase_length;
    for ( length = txinfo->preedit_char_len - ( erase_first + erase_length );
	  length > 0; length--, fb1++, fb2++ )
	*fb1 = *fb2;
    wc1 = txinfo->preedit_char + erase_first;
    wc2 = wc1 + erase_length;
    for ( length = txinfo->preedit_char_len - ( erase_first + erase_length );
	  length > 0; length--, wc1++, wc2++ )
	*wc1 = *wc2;
    *wc1 = 0;
    txinfo->preedit_char_len -= erase_length;
    return;
}

static void
xsf_replace_feedback( ic, pds )
    FXimp_Client		  *ic;
    XIMPreeditDrawCallbackStruct  *pds;
{
    XIMText	*text = pds->text;
    XIMFeedback	*old_fb, *new_fb, *sav_fb, *cpy_fb;
    int		 i;
    int		 length;
#ifndef SVR4
    char	*malloc();
#endif
    if ( !ic->preedit_text.preedit_attr ) {
	ic->preedit_text.preedit_attr = new_fb =
		(XIMFeedback *)malloc( sizeof( XIMFeedback ) * text->length );
	old_fb = text->feedback;
	for ( i = 0; i < (int)text->length; i++, old_fb++, new_fb++ )
	    *new_fb = *old_fb;
	return;
    }
    if ( pds->chg_length < (int)text->length ) {
	length = ic->preedit_text.preedit_char_len
			- pds->chg_length + text->length;
	sav_fb = new_fb =
		    (XIMFeedback *)malloc( sizeof( XIMFeedback ) * length );
	old_fb = ic->preedit_text.preedit_attr;
	for ( i = 0; i < pds->chg_first; i++, old_fb++, new_fb++ )
	    *new_fb = *old_fb;
	cpy_fb = text->feedback;
	for ( i = 0; i < (int)text->length; i++, cpy_fb++, new_fb++ )
	    *new_fb = *cpy_fb;
	old_fb += pds->chg_length;
	length = ic->preedit_text.preedit_char_len
			- (  pds->caret + pds->chg_length );
	for ( ; length > 0; length--, old_fb++, new_fb++ )
	    *new_fb = *old_fb;

	free( ic->preedit_text.preedit_attr );
	ic->preedit_text.preedit_attr = sav_fb;
    } else {
	cpy_fb = text->feedback;
	new_fb = ic->preedit_text.preedit_attr + pds->chg_first;
	for ( i = 0; i < (int)text->length; i++, cpy_fb++, new_fb++ )
	    *new_fb = *cpy_fb;
	cpy_fb = ic->preedit_text.preedit_attr
			+  pds->chg_first + pds->chg_length;
	length = ic->preedit_text.preedit_char_len
			- (  pds->caret + pds->chg_length );
	for ( ; length > 0; length--, cpy_fb++, new_fb++ )
	    *new_fb = *cpy_fb;
    }
    return;
}

static void
xsf_replace_text( ic, pds )
    FXimp_Client		  *ic;
    XIMPreeditDrawCallbackStruct  *pds;
{
    XIMText	*text = pds->text;
    wchar_t	*old_wc, *new_wc, *sav_wc, *cpy_wc;
    wchar_t	*wide_char;
    int		 i;
    int		 text_length;
    int		 length;
#ifndef SVR4
    char	*malloc();
#endif

    text_length = xsf_mbstowcs( pds->text, &wide_char );

    if ( !ic->preedit_text.preedit_char ) {
	ic->preedit_text.preedit_char_len = text_length;
	ic->preedit_text.preedit_char = wide_char;
	return;
    }

    if ( pds->chg_length < text_length ) {
	length = ic->preedit_text.preedit_char_len
			- pds->chg_length + text->length + 1;
	sav_wc = new_wc = (wchar_t *)malloc( sizeof( wchar_t ) * length );
	old_wc = ic->preedit_text.preedit_char;
	for ( i = 0; i < pds->chg_first; i++, old_wc++, new_wc++ )
		*new_wc = *old_wc;
	cpy_wc = wide_char;
	for ( i = 0; i < text_length; i++, cpy_wc++, new_wc++ )
		*new_wc = *cpy_wc;
	old_wc += pds->chg_length;
	length = ic->preedit_text.preedit_char_len
			- (  pds->caret + pds->chg_length );
	for ( ; length > 0; length--, old_wc++, new_wc++ )
		*new_wc = *old_wc;
	*new_wc = 0;
	free( wide_char );
	free( ic->preedit_text.preedit_char );
	ic->preedit_text.preedit_char = sav_wc;
    } else {
	cpy_wc = wide_char;
	new_wc = ic->preedit_text.preedit_char + pds->chg_first;
	for ( i = 0; i < text_length; i++, cpy_wc++, new_wc++ )
		*new_wc = *cpy_wc;
	cpy_wc = ic->preedit_text.preedit_char
			+  pds->chg_first + pds->chg_length;
	length = ic->preedit_text.preedit_char_len
			- (  pds->caret + pds->chg_length );
	for ( ; length > 0; length--, cpy_wc++, new_wc++ )
		*new_wc = *cpy_wc;
	*new_wc = 0;
	free(wide_char);
    }
    ic->preedit_text.preedit_char_len += text->length - pds->chg_length;
    return;
}

/*
 * Convert mbs to wcs when XIMText is mbs.
 */
static int
xsf_mbstowcs( ximtext, textarea )
    XIMText        *ximtext;
    wchar_t        **textarea;
{
    *textarea = NULL ;

    if (!ximtext)
	return ( 0 ) ;
    if ( ( !ximtext->length )
      || ( !ximtext->string.multi_byte ) )
		return ( 0 ) ;
    if ( ximtext->encoding_is_wchar ) {
	int		wcbl = ( ximtext->length + 1 ) * sizeof(wchar_t) ;
	*textarea = (wchar_t *)malloc( wcbl ) ;
	(void)bzero( *textarea, wcbl ) ;
	(void)bcopy( ximtext->string.wide_char,
			*textarea, ximtext->length * sizeof(wchar_t) ) ;
	return ( ximtext->length ) ;
    } else {
	int             mbbl = (ximtext->length + 1) * MB_LEN_MAX;
	*textarea = (wchar_t *) calloc(sizeof(wchar_t), mbbl);

	mbstowcs(*textarea, ximtext->string.multi_byte, mbbl);
	textarea[mbbl] = NULL;
	return ( wslen( *textarea ) ) ;
    }
}

